import os
import os.path
from collections import defaultdict
from functools import partial

import numpy as np

from yt.frontends.art.definitions import (
    hydro_struct,
    particle_fields,
    particle_star_fields,
    star_struct,
)
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.fortran_utils import read_vector, skip
from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.logger import ytLogger as mylog


class IOHandlerART(BaseIOHandler):
    _dataset_type = "art"
    tb, ages = None, None
    cache = None
    masks = None
    caching = True

    def __init__(self, *args, **kwargs):
        self.cache = {}
        self.masks = {}
        super().__init__(*args, **kwargs)
        self.ws = self.ds.parameters["wspecies"]
        self.ls = self.ds.parameters["lspecies"]
        self.file_particle = self.ds._file_particle_data
        self.file_stars = self.ds._file_particle_stars
        self.Nrow = self.ds.parameters["Nrow"]

    def _read_fluid_selection(self, chunks, selector, fields, size):
        # Chunks in this case will have affiliated domain subset objects
        # Each domain subset will contain a hydro_offset array, which gives
        # pointers to level-by-level hydro information
        tr = defaultdict(list)
        cp = 0
        for chunk in chunks:
            for subset in chunk.objs:
                # Now we read the entire thing
                f = open(subset.domain.ds._file_amr, "rb")
                # This contains the boundary information, so we skim through
                # and pick off the right vectors
                rv = subset.fill(f, fields, selector)
                for ft, f in fields:
                    d = rv.pop(f)
                    mylog.debug(
                        "Filling %s with %s (%0.3e %0.3e) (%s:%s)",
                        f,
                        d.size,
                        d.min(),
                        d.max(),
                        cp,
                        cp + d.size,
                    )
                    tr[(ft, f)].append(d)
                cp += d.size
        d = {}
        for field in fields:
            d[field] = np.concatenate(tr.pop(field))
        return d

    def _get_mask(self, selector, ftype):
        key = (selector, ftype)
        if key in self.masks.keys() and self.caching:
            return self.masks[key]
        pstr = "particle_position_%s"
        x, y, z = (self._get_field((ftype, pstr % ax)) for ax in "xyz")
        mask = selector.select_points(x, y, z, 0.0)
        if self.caching:
            self.masks[key] = mask
            return self.masks[key]
        else:
            return mask

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        for _chunk in chunks:
            for ptype in sorted(ptf):
                x = self._get_field((ptype, "particle_position_x"))
                y = self._get_field((ptype, "particle_position_y"))
                z = self._get_field((ptype, "particle_position_z"))
                yield ptype, (x, y, z)

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        for _chunk in chunks:
            for ptype, field_list in sorted(ptf.items()):
                x = self._get_field((ptype, "particle_position_x"))
                y = self._get_field((ptype, "particle_position_y"))
                z = self._get_field((ptype, "particle_position_z"))
                mask = selector.select_points(x, y, z, 0.0)
                if mask is None:
                    continue
                for field in field_list:
                    data = self._get_field((ptype, field))
                    yield (ptype, field), data[mask]

    def _get_field(self, field):
        if field in self.cache.keys() and self.caching:
            mylog.debug("Cached %s", str(field))
            return self.cache[field]
        mylog.debug("Reading %s", str(field))
        tr = {}
        ftype, fname = field
        ptmax = self.ws[-1]
        pbool, idxa, idxb = _determine_field_size(self.ds, ftype, self.ls, ptmax)
        npa = idxb - idxa
        sizes = np.diff(np.concatenate(([0], self.ls)))
        rp = partial(
            read_particles, self.file_particle, self.Nrow, idxa=idxa, idxb=idxb
        )
        for ax in "xyz":
            if fname.startswith(f"particle_position_{ax}"):
                dd = self.ds.domain_dimensions[0]
                off = 1.0 / dd
                tr[field] = rp(fields=[ax])[0] / dd - off
            if fname.startswith(f"particle_velocity_{ax}"):
                (tr[field],) = rp(fields=["v" + ax])
        if fname.startswith("particle_mass"):
            a = 0
            data = np.zeros(npa, dtype="f8")
            for ptb, size, m in zip(pbool, sizes, self.ws):
                if ptb:
                    data[a : a + size] = m
                    a += size
            tr[field] = data
        elif fname == "particle_index":
            tr[field] = np.arange(idxa, idxb)
        elif fname == "particle_type":
            a = 0
            data = np.zeros(npa, dtype="int64")
            for i, (ptb, size) in enumerate(zip(pbool, sizes)):
                if ptb:
                    data[a : a + size] = i
                    a += size
            tr[field] = data
        if pbool[-1] and fname in particle_star_fields:
            data = read_star_field(self.file_stars, field=fname)
            temp = tr.get(field, np.zeros(npa, "f8"))
            nstars = self.ls[-1] - self.ls[-2]
            if nstars > 0:
                temp[-nstars:] = data
            tr[field] = temp
        if fname == "particle_creation_time":
            self.tb, self.ages, data = interpolate_ages(
                tr[field][-nstars:],
                self.file_stars,
                self.tb,
                self.ages,
                self.ds.current_time,
            )
            temp = tr.get(field, np.zeros(npa, "f8"))
            temp[-nstars:] = data
            tr[field] = temp
            del data
        # We check again, after it's been filled
        if fname.startswith("particle_mass"):
            # We now divide by NGrid in order to make this match up.  Note that
            # this means that even when requested in *code units*, we are
            # giving them as modified by the ng value.  This only works for
            # dark_matter -- stars are regular matter.
            tr[field] /= self.ds.domain_dimensions.prod()
        if tr == {}:
            tr = {f: np.array([]) for f in [field]}
        if self.caching:
            self.cache[field] = tr[field]
            return self.cache[field]
        else:
            return tr[field]


class IOHandlerDarkMatterART(IOHandlerART):
    _dataset_type = "dm_art"

    def _count_particles(self, data_file):
        return {
            k: self.ds.parameters["lspecies"][i]
            for i, k in enumerate(self.ds.particle_types_raw)
        }

    def _identify_fields(self, domain):
        field_list = []
        self.particle_field_list = [f for f in particle_fields]
        for ptype in self.ds.particle_types_raw:
            for pfield in self.particle_field_list:
                pfn = (ptype, pfield)
                field_list.append(pfn)
        return field_list, {}

    def _get_field(self, field):
        if field in self.cache.keys() and self.caching:
            mylog.debug("Cached %s", str(field))
            return self.cache[field]
        mylog.debug("Reading %s", str(field))
        tr = {}
        ftype, fname = field
        ptmax = self.ws[-1]
        pbool, idxa, idxb = _determine_field_size(self.ds, ftype, self.ls, ptmax)
        npa = idxb - idxa
        sizes = np.diff(np.concatenate(([0], self.ls)))
        rp = partial(
            read_particles, self.file_particle, self.Nrow, idxa=idxa, idxb=idxb
        )
        for ax in "xyz":
            if fname.startswith(f"particle_position_{ax}"):
                # This is not the same as domain_dimensions
                dd = self.ds.parameters["ng"]
                off = 1.0 / dd
                tr[field] = rp(fields=[ax])[0] / dd - off
            if fname.startswith(f"particle_velocity_{ax}"):
                (tr[field],) = rp(["v" + ax])
        if fname.startswith("particle_mass"):
            a = 0
            data = np.zeros(npa, dtype="f8")
            for ptb, size, m in zip(pbool, sizes, self.ws):
                if ptb:
                    data[a : a + size] = m
                    a += size
            tr[field] = data
        elif fname == "particle_index":
            tr[field] = np.arange(idxa, idxb)
        elif fname == "particle_type":
            a = 0
            data = np.zeros(npa, dtype="int64")
            for i, (ptb, size) in enumerate(zip(pbool, sizes)):
                if ptb:
                    data[a : a + size] = i
                    a += size
            tr[field] = data
        # We check again, after it's been filled
        if fname.startswith("particle_mass"):
            # We now divide by NGrid in order to make this match up.  Note that
            # this means that even when requested in *code units*, we are
            # giving them as modified by the ng value.  This only works for
            # dark_matter -- stars are regular matter.
            tr[field] /= self.ds.domain_dimensions.prod()
        if tr == {}:
            tr[field] = np.array([])
        if self.caching:
            self.cache[field] = tr[field]
            return self.cache[field]
        else:
            return tr[field]

    def _yield_coordinates(self, data_file):
        for ptype in self.ds.particle_types_raw:
            x = self._get_field((ptype, "particle_position_x"))
            y = self._get_field((ptype, "particle_position_y"))
            z = self._get_field((ptype, "particle_position_z"))

            yield ptype, np.stack((x, y, z), axis=-1)


def _determine_field_size(pf, field, lspecies, ptmax):
    pbool = np.zeros(len(lspecies), dtype="bool")
    idxas = np.concatenate(
        (
            [
                0,
            ],
            lspecies[:-1],
        )
    )
    idxbs = lspecies
    if "specie" in field:
        index = int(field.replace("specie", ""))
        pbool[index] = True
    else:
        raise RuntimeError
    idxa, idxb = idxas[pbool][0], idxbs[pbool][-1]
    return pbool, idxa, idxb


def interpolate_ages(
    data, file_stars, interp_tb=None, interp_ages=None, current_time=None
):
    if interp_tb is None:
        t_stars, a_stars = read_star_field(file_stars, field="t_stars")
        # timestamp of file should match amr timestamp
        if current_time:
            tdiff = YTQuantity(b2t(t_stars), "Gyr") - current_time.in_units("Gyr")
            if np.abs(tdiff) > 1e-4:
                mylog.info("Timestamp mismatch in star particle header: %s", tdiff)
        mylog.info("Interpolating ages")
        interp_tb, interp_ages = b2t(data)
        interp_tb = YTArray(interp_tb, "Gyr")
        interp_ages = YTArray(interp_ages, "Gyr")
    temp = np.interp(data, interp_tb, interp_ages)
    return interp_tb, interp_ages, temp


def _read_art_level_info(
    f, level_oct_offsets, level, coarse_grid=128, ncell0=None, root_level=None
):
    pos = f.tell()
    f.seek(level_oct_offsets[level])
    # Get the info for this level, skip the rest
    junk, nLevel, iOct = read_vector(f, "i", ">")

    # fortran indices start at 1

    # Skip all the oct index data
    le = np.zeros((nLevel, 3), dtype="int64")
    fl = np.ones((nLevel, 6), dtype="int64")
    iocts = np.zeros(nLevel + 1, dtype="int64")
    idxa, idxb = 0, 0
    chunk = int(1e6)  # this is ~111MB for 15 dimensional 64 bit arrays
    left = nLevel
    while left > 0:
        this_chunk = min(chunk, left)
        idxb = idxa + this_chunk
        data = np.fromfile(f, dtype=">i", count=this_chunk * 15)
        data = data.reshape(this_chunk, 15)
        left -= this_chunk
        le[idxa:idxb, :] = data[:, 1:4]
        fl[idxa:idxb, 1] = np.arange(idxa, idxb)
        # pad byte is last, LL2, then ioct right before it
        iocts[idxa:idxb] = data[:, -3]
        idxa = idxa + this_chunk
    del data

    # emulate fortran code
    #     do ic1 = 1 , nLevel
    #       read(19) (iOctPs(i,iOct),i=1,3),(iOctNb(i,iOct),i=1,6),
    # &                iOctPr(iOct), iOctLv(iOct), iOctLL1(iOct),
    # &                iOctLL2(iOct)
    #       iOct = iOctLL1(iOct)

    # ioct always represents the index of the next variable
    # not the current, so shift forward one index
    # the last index isn't used
    iocts[1:] = iocts[:-1]  # shift
    iocts = iocts[:nLevel]  # chop off the last, unused, index
    iocts[0] = iOct  # starting value

    # now correct iocts for fortran indices start @ 1
    iocts = iocts - 1

    assert np.unique(iocts).shape[0] == nLevel

    # left edges are expressed as if they were on
    # level 15, so no matter what level max(le)=2**15
    # correct to the yt convention
    # le = le/2**(root_level-1-level)-1

    # try to find the root_level first
    def cfc(root_level, level, le):
        d_x = 1.0 / (2.0 ** (root_level - level + 1))
        fc = (d_x * le) - 2 ** (level - 1)
        return fc

    if root_level is None:
        root_level = np.floor(np.log2(le.max() * 1.0 / coarse_grid))
        root_level = root_level.astype("int64")
        for _ in range(10):
            fc = cfc(root_level, level, le)
            go = np.diff(np.unique(fc)).min() < 1.1
            if go:
                break
            root_level += 1
    else:
        fc = cfc(root_level, level, le)
    unitary_center = fc / (coarse_grid * 2.0 ** (level - 1))
    assert np.all(unitary_center < 1.0)

    # again emulate the fortran code
    # This is all for calculating child oct locations
    # iC_ = iC + nbshift
    # iO = ishft ( iC_ , - ndim )
    # id = ishft ( 1, MaxLevel - iOctLv(iO) )
    # j  = iC_ + 1 - ishft( iO , ndim )
    # Posx   = d_x * (iOctPs(1,iO) + sign ( id , idelta(j,1) ))
    # Posy   = d_x * (iOctPs(2,iO) + sign ( id , idelta(j,2) ))
    # Posz   = d_x * (iOctPs(3,iO) + sign ( id , idelta(j,3) ))
    # idelta = [[-1,  1, -1,  1, -1,  1, -1,  1],
    #           [-1, -1,  1,  1, -1, -1,  1,  1],
    #           [-1, -1, -1, -1,  1,  1,  1,  1]]
    # idelta = np.array(idelta)
    # if ncell0 is None:
    #     ncell0 = coarse_grid**3
    # nchild = 8
    # ndim = 3
    # nshift = nchild -1
    # nbshift = nshift - ncell0
    # iC = iocts #+ nbshift
    # iO = iC >> ndim #possibly >>
    # id = 1 << (root_level - level)
    # j = iC + 1 - ( iO << 3)
    # delta = np.abs(id)*idelta[:,j-1]

    # try without the -1
    # le = le/2**(root_level+1-level)
    # now read the hvars and vars arrays
    # we are looking for iOctCh
    # we record if iOctCh is >0, in which it is subdivided
    # iOctCh  = np.zeros((nLevel+1,8),dtype='bool')
    f.seek(pos)
    return unitary_center, fl, iocts, nLevel, root_level


def get_ranges(
    skip, count, field, words=6, real_size=4, np_per_page=4096 ** 2, num_pages=1
):
    # translate every particle index into a file position ranges
    ranges = []
    arr_size = np_per_page * real_size
    idxa, idxb = 0, 0
    posa, posb = 0, 0
    for _page in range(num_pages):
        idxb += np_per_page
        for i, fname in enumerate(["x", "y", "z", "vx", "vy", "vz"]):
            posb += arr_size
            if i == field or fname == field:
                if skip < np_per_page and count > 0:
                    left_in_page = np_per_page - skip
                    this_count = min(left_in_page, count)
                    count -= this_count
                    start = posa + skip * real_size
                    end = posa + this_count * real_size
                    ranges.append((start, this_count))
                    skip = 0
                    assert end <= posb
                else:
                    skip -= np_per_page
            posa += arr_size
        idxa += np_per_page
    assert count == 0
    return ranges


def read_particles(file, Nrow, idxa, idxb, fields):
    words = 6  # words (reals) per particle: x,y,z,vx,vy,vz
    real_size = 4  # for file_particle_data; not always true?
    np_per_page = Nrow ** 2  # defined in ART a_setup.h, # of particles/page
    num_pages = os.path.getsize(file) // (real_size * words * np_per_page)
    fh = open(file)
    skip, count = idxa, idxb - idxa
    kwargs = dict(
        words=words, real_size=real_size, np_per_page=np_per_page, num_pages=num_pages
    )
    arrs = []
    for field in fields:
        ranges = get_ranges(skip, count, field, **kwargs)
        data = None
        for seek, this_count in ranges:
            fh.seek(seek)
            temp = np.fromfile(fh, count=this_count, dtype=">f4")
            if data is None:
                data = temp
            else:
                data = np.concatenate((data, temp))
        arrs.append(data.astype("f8"))
    fh.close()
    return arrs


def read_star_field(file, field=None):
    data = {}
    with open(file, "rb") as fh:
        for dtype, variables in star_struct:
            found = (
                isinstance(variables, tuple) and field in variables
            ) or field == variables
            if found:
                data[field] = read_vector(fh, dtype[1], dtype[0])
            else:
                skip(fh, endian=">")
    return data.pop(field)


def _read_child_mask_level(f, level_child_offsets, level, nLevel, nhydro_vars):
    f.seek(level_child_offsets[level])
    ioctch = np.zeros(nLevel, dtype="uint8")
    idc = np.zeros(nLevel, dtype="int32")

    chunk = int(1e6)
    left = nLevel
    width = nhydro_vars + 6
    a, b = 0, 0
    while left > 0:
        chunk = min(chunk, left)
        b += chunk
        arr = np.fromfile(f, dtype=">i", count=chunk * width)
        arr = arr.reshape((width, chunk), order="F")
        assert np.all(arr[0, :] == arr[-1, :])  # pads must be equal
        idc[a:b] = arr[1, :] - 1  # fix fortran indexing
        ioctch[a:b] = arr[2, :] == 0  # if it is above zero, then refined available
        # zero in the mask means there is refinement available
        a = b
        left -= chunk
    assert left == 0
    return idc, ioctch


nchem = 8 + 2
dtyp = np.dtype(f">i4,>i8,>i8,>{nchem}f4,>2f4,>i4")


def _read_child_level(
    f,
    level_child_offsets,
    level_oct_offsets,
    level_info,
    level,
    fields,
    domain_dimensions,
    ncell0,
    nhydro_vars=10,
    nchild=8,
    noct_range=None,
):
    # emulate the fortran code for reading cell data
    # read ( 19 ) idc, iOctCh(idc), (hvar(i,idc),i=1,nhvar),
    #    &                 (var(i,idc), i=2,3)
    # contiguous 8-cell sections are for the same oct;
    # ie, we don't write out just the 0 cells, then the 1 cells
    # optionally, we only read noct_range to save memory
    left_index, fl, octs, nocts, root_level = _read_art_level_info(
        f, level_oct_offsets, level, coarse_grid=domain_dimensions[0]
    )
    if noct_range is None:
        nocts = level_info[level]
        ncells = nocts * 8
        f.seek(level_child_offsets[level])
        arr = np.fromfile(f, dtype=hydro_struct, count=ncells)
        assert np.all(arr["pad1"] == arr["pad2"])  # pads must be equal
        # idc = np.argsort(arr['idc']) #correct fortran indices
        # translate idc into icell, and then to iOct
        icell = (arr["idc"] >> 3) << 3
        iocts = (icell - ncell0) / nchild  # without a F correction, theres a +1
        # assert that the children are read in the same order as the octs
        assert np.all(octs == iocts[::nchild])
    else:
        start, end = noct_range
        nocts = min(end - start, level_info[level])
        end = start + nocts
        ncells = nocts * 8
        skip = np.dtype(hydro_struct).itemsize * start * 8
        f.seek(level_child_offsets[level] + skip)
        arr = np.fromfile(f, dtype=hydro_struct, count=ncells)
        assert np.all(arr["pad1"] == arr["pad2"])  # pads must be equal
    source = {}
    for field in fields:
        sh = (nocts, 8)
        source[field] = np.reshape(arr[field], sh, order="C").astype("float64")
    return source


def _read_root_level(f, level_offsets, level_info, nhydro_vars=10):
    nocts = level_info[0]
    f.seek(level_offsets[0])  # Ditch the header
    hvar = read_vector(f, "f", ">")
    var = read_vector(f, "f", ">")
    hvar = hvar.reshape((nhydro_vars, nocts * 8), order="F")
    var = var.reshape((2, nocts * 8), order="F")
    arr = np.concatenate((hvar, var))
    return arr


# All of these functions are to convert from hydro time var to
# proper time
sqrt = np.sqrt
sign = np.sign


def find_root(f, a, b, tol=1e-6):
    c = (a + b) / 2.0
    last = -np.inf
    assert sign(f(a)) != sign(f(b))
    while np.abs(f(c) - last) > tol:
        last = f(c)
        if sign(last) == sign(f(b)):
            b = c
        else:
            a = c
        c = (a + b) / 2.0
    return c


def quad(fintegrand, xmin, xmax, n=1e4):
    spacings = np.logspace(np.log10(xmin), np.log10(xmax), num=int(n))
    integrand_arr = fintegrand(spacings)
    val = np.trapz(integrand_arr, dx=np.diff(spacings))
    return val


def a2b(at, Om0=0.27, Oml0=0.73, h=0.700):
    def f_a2b(x):
        val = 0.5 * sqrt(Om0) / x ** 3.0
        val /= sqrt(Om0 / x ** 3.0 + Oml0 + (1.0 - Om0 - Oml0) / x ** 2.0)
        return val

    # val, err = si.quad(f_a2b,1,at)
    val = quad(f_a2b, 1, at)
    return val


def b2a(bt, **kwargs):
    # converts code time into expansion factor
    # if Om0 ==1and OmL == 0 then b2a is (1 / (1-td))**2
    # if bt < -190.0 or bt > -.10:  raise 'bt outside of range'
    def f_b2a(at):
        return a2b(at, **kwargs) - bt

    return find_root(f_b2a, 1e-4, 1.1)
    # return so.brenth(f_b2a,1e-4,1.1)
    # return brent.brent(f_b2a)


def a2t(at, Om0=0.27, Oml0=0.73, h=0.700):
    def integrand(x):
        return 1.0 / (x * sqrt(Oml0 + Om0 * x ** -3.0))

    # current_time,err = si.quad(integrand,0.0,at,epsabs=1e-6,epsrel=1e-6)
    current_time = quad(integrand, 1e-4, at)
    # spacings = np.logspace(-5,np.log10(at),num=int(1e5))
    # integrand_arr = integrand(spacings)
    # current_time = np.trapz(integrand_arr,dx=np.diff(spacings))
    current_time *= 9.779 / h
    return current_time


def b2t(tb, n=1e2, logger=None, **kwargs):
    tb = np.array(tb)
    if isinstance(tb, type(1.1)):
        return a2t(b2a(tb))
    if tb.shape == ():
        return a2t(b2a(tb))
    if len(tb) < n:
        n = len(tb)
    tbs = -1.0 * np.logspace(np.log10(-tb.min()), np.log10(-tb.max()), n)
    ages = []
    for i, tbi in enumerate(tbs):
        ages += (a2t(b2a(tbi)),)
        if logger:
            logger(i)
    ages = np.array(ages)
    return tbs, ages



from io import BytesIO

import matplotlib
import numpy as np
from packaging.version import parse as parse_version

from yt.funcs import (
    get_brewer_cmap,
    get_interactivity,
    is_sequence,
    matplotlib_style_context,
    mylog,
)

from ._commons import get_canvas, validate_image_name

BACKEND_SPECS = {
    "GTK": ["backend_gtk", "FigureCanvasGTK", "FigureManagerGTK"],
    "GTKAgg": ["backend_gtkagg", "FigureCanvasGTKAgg", None],
    "GTKCairo": ["backend_gtkcairo", "FigureCanvasGTKCairo", None],
    "MacOSX": ["backend_macosx", "FigureCanvasMac", "FigureManagerMac"],
    "Qt4Agg": ["backend_qt4agg", "FigureCanvasQTAgg", None],
    "Qt5Agg": ["backend_qt5agg", "FigureCanvasQTAgg", None],
    "TkAgg": ["backend_tkagg", "FigureCanvasTkAgg", None],
    "WX": ["backend_wx", "FigureCanvasWx", None],
    "WXAgg": ["backend_wxagg", "FigureCanvasWxAgg", None],
    "GTK3Cairo": [
        "backend_gtk3cairo",
        "FigureCanvasGTK3Cairo",
        "FigureManagerGTK3Cairo",
    ],
    "GTK3Agg": ["backend_gtk3agg", "FigureCanvasGTK3Agg", "FigureManagerGTK3Agg"],
    "WebAgg": ["backend_webagg", "FigureCanvasWebAgg", None],
    "nbAgg": ["backend_nbagg", "FigureCanvasNbAgg", "FigureManagerNbAgg"],
    "agg": ["backend_agg", "FigureCanvasAgg", None],
}


class CallbackWrapper:
    def __init__(self, viewer, window_plot, frb, field, font_properties, font_color):
        self.frb = frb
        self.data = frb.data_source
        self._axes = window_plot.axes
        self._figure = window_plot.figure
        if len(self._axes.images) > 0:
            self.image = self._axes.images[0]
        if frb.axis < 3:
            DD = frb.ds.domain_width
            xax = frb.ds.coordinates.x_axis[frb.axis]
            yax = frb.ds.coordinates.y_axis[frb.axis]
            self._period = (DD[xax], DD[yax])
        self.ds = frb.ds
        self.xlim = viewer.xlim
        self.ylim = viewer.ylim
        self._axes_unit_names = viewer._axes_unit_names
        if "OffAxisSlice" in viewer._plot_type:
            self._type_name = "CuttingPlane"
        else:
            self._type_name = viewer._plot_type
        self.aspect = window_plot._aspect
        self.font_properties = font_properties
        self.font_color = font_color
        self.field = field


class PlotMPL:
    """A base class for all yt plots made using matplotlib, that is backend independent."""

    def __init__(self, fsize, axrect, figure, axes):
        """Initialize PlotMPL class"""
        import matplotlib.figure

        self._plot_valid = True
        if figure is None:
            if not is_sequence(fsize):
                fsize = (fsize, fsize)
            self.figure = matplotlib.figure.Figure(figsize=fsize, frameon=True)
        else:
            figure.set_size_inches(fsize)
            self.figure = figure
        if axes is None:
            self._create_axes(axrect)
        else:
            axes.cla()
            axes.set_position(axrect)
            self.axes = axes
        self.interactivity = get_interactivity()

        figure_canvas, figure_manager = self._get_canvas_classes()
        self.canvas = figure_canvas(self.figure)
        if figure_manager is not None:
            self.manager = figure_manager(self.canvas, 1)

        for which in ["major", "minor"]:
            for axis in "xy":
                self.axes.tick_params(
                    which=which, axis=axis, direction="in", top=True, right=True
                )

    def _create_axes(self, axrect):
        self.axes = self.figure.add_axes(axrect)

    def _get_canvas_classes(self):

        if self.interactivity:
            key = str(matplotlib.get_backend())
        else:
            key = "agg"

        try:
            module, fig_canvas, fig_manager = BACKEND_SPECS[key]
        except KeyError:
            return

        mod = __import__(
            "matplotlib.backends",
            globals(),
            locals(),
            [module],
            0,
        )
        submod = getattr(mod, module)
        FigureCanvas = getattr(submod, fig_canvas)
        if fig_manager is not None:
            FigureManager = getattr(submod, fig_manager)
            return FigureCanvas, FigureManager

        return FigureCanvas, None

    def save(self, name, mpl_kwargs=None, canvas=None):
        """Choose backend and save image to disk"""

        if mpl_kwargs is None:
            mpl_kwargs = {}
        if "papertype" not in mpl_kwargs and parse_version(
            matplotlib.__version__
        ) < parse_version("3.3.0"):
            mpl_kwargs["papertype"] = "auto"

        name = validate_image_name(name)

        try:
            canvas = get_canvas(self.figure, name)
        except ValueError:
            canvas = self.canvas

        mylog.info("Saving plot %s", name)
        with matplotlib_style_context():
            canvas.print_figure(name, **mpl_kwargs)
        return name

    def show(self):
        try:
            self.manager.show()
        except AttributeError:
            self.canvas.show()

    def _get_labels(self):
        ax = self.axes
        labels = ax.xaxis.get_ticklabels() + ax.yaxis.get_ticklabels()
        labels += ax.xaxis.get_minorticklabels()
        labels += ax.yaxis.get_minorticklabels()
        labels += [
            ax.title,
            ax.xaxis.label,
            ax.yaxis.label,
            ax.xaxis.get_offset_text(),
            ax.yaxis.get_offset_text(),
        ]
        return labels

    def _set_font_properties(self, font_properties, font_color):
        for label in self._get_labels():
            label.set_fontproperties(font_properties)
            if font_color is not None:
                label.set_color(self.font_color)

    def _repr_png_(self):
        from ._mpl_imports import FigureCanvasAgg

        canvas = FigureCanvasAgg(self.figure)
        f = BytesIO()
        with matplotlib_style_context():
            canvas.print_figure(f)
        f.seek(0)
        return f.read()


class ImagePlotMPL(PlotMPL):
    """A base class for yt plots made using imshow"""

    def __init__(self, fsize, axrect, caxrect, zlim, figure, axes, cax):
        """Initialize ImagePlotMPL class object"""
        super().__init__(fsize, axrect, figure, axes)
        self.zmin, self.zmax = zlim
        if cax is None:
            self.cax = self.figure.add_axes(caxrect)
        else:
            cax.cla()
            cax.set_position(caxrect)
            self.cax = cax

    def _init_image(self, data, cbnorm, cblinthresh, cmap, extent, aspect):
        """Store output of imshow in image variable"""
        cbnorm_kwargs = dict(
            vmin=float(self.zmin) if self.zmin is not None else None,
            vmax=float(self.zmax) if self.zmax is not None else None,
        )
        if cbnorm == "log10":
            cbnorm_cls = matplotlib.colors.LogNorm
        elif cbnorm == "linear":
            cbnorm_cls = matplotlib.colors.Normalize
        elif cbnorm == "symlog":
            # if cblinthresh is not specified, try to come up with a reasonable default
            vmin = float(np.nanmin(data))
            vmax = float(np.nanmax(data))
            if cblinthresh is None:
                cblinthresh = np.nanmin(np.absolute(data)[data != 0])

            cbnorm_kwargs.update(dict(linthresh=cblinthresh, vmin=vmin, vmax=vmax))
            MPL_VERSION = parse_version(matplotlib.__version__)
            if MPL_VERSION >= parse_version("3.2.0"):
                # note that this creates an inconsistency between mpl versions
                # since the default value previous to mpl 3.4.0 is np.e
                # but it is only exposed since 3.2.0
                cbnorm_kwargs["base"] = 10

            cbnorm_cls = matplotlib.colors.SymLogNorm
        else:
            raise ValueError(f"Unknown value `cbnorm` == {cbnorm}")

        norm = cbnorm_cls(**cbnorm_kwargs)

        extent = [float(e) for e in extent]
        # tuple colormaps are from palettable (or brewer2mpl)
        if isinstance(cmap, tuple):
            cmap = get_brewer_cmap(cmap)

        if self._transform is None:
            # sets the transform to be an ax.TransData object, where the
            # coordiante system of the data is controlled by the xlim and ylim
            # of the data.
            transform = self.axes.transData
        else:
            transform = self._transform
        if hasattr(self.axes, "set_extent"):
            # CartoPy hangs if we do not set_extent before imshow if we are
            # displaying a small subset of the globe.  What I believe happens is
            # that the transform for the points on the outside results in
            # infinities, and then the scipy.spatial cKDTree hangs trying to
            # identify nearest points.
            #
            # Also, set_extent is defined by cartopy, so not all axes will have
            # it as a method.
            #
            # A potential downside is that other images may change, but I believe
            # the result of imshow is to set_extent *regardless*.  This just
            # changes the order in which it happens.
            #
            # NOTE: This is currently commented out because it breaks in some
            # instances.  It is left as a historical note because we will
            # eventually need some form of it.
            # self.axes.set_extent(extent)
            pass
        self.image = self.axes.imshow(
            data.to_ndarray(),
            origin="lower",
            extent=extent,
            norm=norm,
            aspect=aspect,
            cmap=cmap,
            interpolation="nearest",
            transform=transform,
        )
        if cbnorm == "symlog":
            formatter = matplotlib.ticker.LogFormatterMathtext(linthresh=cblinthresh)
            self.cb = self.figure.colorbar(self.image, self.cax, format=formatter)
            if np.nanmin(data) >= 0.0:
                yticks = [np.nanmin(data).v] + list(
                    10
                    ** np.arange(
                        np.rint(np.log10(cblinthresh)),
                        np.ceil(np.log10(np.nanmax(data))) + 1,
                    )
                )
            elif np.nanmax(data) <= 0.0:
                yticks = (
                    list(
                        -(
                            10
                            ** np.arange(
                                np.floor(np.log10(-np.nanmin(data))),
                                np.rint(np.log10(cblinthresh)) - 1,
                                -1,
                            )
                        )
                    )
                    + [np.nanmax(data).v]
                )
            else:
                yticks = (
                    list(
                        -(
                            10
                            ** np.arange(
                                np.floor(np.log10(-np.nanmin(data))),
                                np.rint(np.log10(cblinthresh)) - 1,
                                -1,
                            )
                        )
                    )
                    + [0]
                    + list(
                        10
                        ** np.arange(
                            np.rint(np.log10(cblinthresh)),
                            np.ceil(np.log10(np.nanmax(data))) + 1,
                        )
                    )
                )
            self.cb.set_ticks(yticks)
        else:
            self.cb = self.figure.colorbar(self.image, self.cax)
        for which in ["major", "minor"]:
            self.cax.tick_params(which=which, axis="y", direction="in")

    def _get_best_layout(self):

        # Ensure the figure size along the long axis is always equal to _figure_size
        if is_sequence(self._figure_size):
            x_fig_size = self._figure_size[0]
            y_fig_size = self._figure_size[1]
        else:
            x_fig_size = self._figure_size
            y_fig_size = self._figure_size / self._aspect

        if hasattr(self, "_unit_aspect"):
            y_fig_size = y_fig_size * self._unit_aspect

        if self._draw_colorbar:
            cb_size = self._cb_size
            cb_text_size = self._ax_text_size[1] + 0.45
        else:
            cb_size = x_fig_size * 0.04
            cb_text_size = 0.0

        if self._draw_axes:
            x_axis_size = self._ax_text_size[0]
            y_axis_size = self._ax_text_size[1]
        else:
            x_axis_size = x_fig_size * 0.04
            y_axis_size = y_fig_size * 0.04

        top_buff_size = self._top_buff_size

        if not self._draw_axes and not self._draw_colorbar:
            x_axis_size = 0.0
            y_axis_size = 0.0
            cb_size = 0.0
            cb_text_size = 0.0
            top_buff_size = 0.0

        xbins = np.array([x_axis_size, x_fig_size, cb_size, cb_text_size])
        ybins = np.array([y_axis_size, y_fig_size, top_buff_size])

        size = [xbins.sum(), ybins.sum()]

        x_frac_widths = xbins / size[0]
        y_frac_widths = ybins / size[1]

        # axrect is the rectangle defining the area of the
        # axis object of the plot.  Its range goes from 0 to 1 in
        # x and y directions.  The first two values are the x,y
        # start values of the axis object (lower left corner), and the
        # second two values are the size of the axis object.  To get
        # the upper right corner, add the first x,y to the second x,y.
        axrect = (
            x_frac_widths[0],
            y_frac_widths[0],
            x_frac_widths[1],
            y_frac_widths[1],
        )

        # caxrect is the rectangle defining the area of the colorbar
        # axis object of the plot.  It is defined just as the axrect
        # tuple is.
        caxrect = (
            x_frac_widths[0] + x_frac_widths[1],
            y_frac_widths[0],
            x_frac_widths[2],
            y_frac_widths[1],
        )

        return size, axrect, caxrect

    def _toggle_axes(self, choice, draw_frame=None):
        """
        Turn on/off displaying the axis ticks and labels for a plot.

        Parameters
        ----------
        choice : boolean
            If True, set the axes to be drawn. If False, set the axes to not be
            drawn.
        """
        if draw_frame is None:
            draw_frame = choice
        self._draw_axes = choice
        self._draw_frame = draw_frame
        self.axes.set_frame_on(draw_frame)
        self.axes.get_xaxis().set_visible(choice)
        self.axes.get_yaxis().set_visible(choice)
        size, axrect, caxrect = self._get_best_layout()
        self.axes.set_position(axrect)
        self.cax.set_position(caxrect)
        self.figure.set_size_inches(*size)

    def _toggle_colorbar(self, choice):
        """
        Turn on/off displaying the colorbar for a plot

        choice = True or False
        """
        self._draw_colorbar = choice
        self.cax.set_visible(choice)
        size, axrect, caxrect = self._get_best_layout()
        self.axes.set_position(axrect)
        self.cax.set_position(caxrect)
        self.figure.set_size_inches(*size)

    def _get_labels(self):
        labels = super()._get_labels()
        cbax = self.cb.ax
        labels += cbax.yaxis.get_ticklabels()
        labels += [cbax.yaxis.label, cbax.yaxis.get_offset_text()]
        return labels

    def hide_axes(self, draw_frame=None):
        """
        Hide the axes for a plot including ticks and labels
        """
        self._toggle_axes(False, draw_frame)
        return self

    def show_axes(self):
        """
        Show the axes for a plot including ticks and labels
        """
        self._toggle_axes(True)
        return self

    def hide_colorbar(self):
        """
        Hide the colorbar for a plot including ticks and labels
        """
        self._toggle_colorbar(False)
        return self

    def show_colorbar(self):
        """
        Show the colorbar for a plot including ticks and labels
        """
        self._toggle_colorbar(True)
        return self


def get_multi_plot(nx, ny, colorbar="vertical", bw=4, dpi=300, cbar_padding=0.4):
    r"""Construct a multiple axes plot object, with or without a colorbar, into
    which multiple plots may be inserted.

    This will create a set of :class:`matplotlib.axes.Axes`, all lined up into
    a grid, which are then returned to the user and which can be used to plot
    multiple plots on a single figure.

    Parameters
    ----------
    nx : int
        Number of axes to create along the x-direction
    ny : int
        Number of axes to create along the y-direction
    colorbar : {'vertical', 'horizontal', None}, optional
        Should Axes objects for colorbars be allocated, and if so, should they
        correspond to the horizontal or vertical set of axes?
    bw : number
        The base height/width of an axes object inside the figure, in inches
    dpi : number
        The dots per inch fed into the Figure instantiation

    Returns
    -------
    fig : :class:`matplotlib.figure.Figure`
        The figure created inside which the axes reside
    tr : list of list of :class:`matplotlib.axes.Axes` objects
        This is a list, where the inner list is along the x-axis and the outer
        is along the y-axis
    cbars : list of :class:`matplotlib.axes.Axes` objects
        Each of these is an axes onto which a colorbar can be placed.

    Notes
    -----
    This is a simple implementation for a common use case.  Viewing the source
    can be instructive, and is encouraged to see how to generate more
    complicated or more specific sets of multiplots for your own purposes.
    """
    import matplotlib.figure

    hf, wf = 1.0 / ny, 1.0 / nx
    fudge_x = fudge_y = 1.0
    if colorbar is None:
        fudge_x = fudge_y = 1.0
    elif colorbar.lower() == "vertical":
        fudge_x = nx / (cbar_padding + nx)
        fudge_y = 1.0
    elif colorbar.lower() == "horizontal":
        fudge_x = 1.0
        fudge_y = ny / (cbar_padding + ny)
    fig = matplotlib.figure.Figure((bw * nx / fudge_x, bw * ny / fudge_y), dpi=dpi)
    from ._mpl_imports import FigureCanvasAgg

    fig.set_canvas(FigureCanvasAgg(fig))
    fig.subplots_adjust(
        wspace=0.0, hspace=0.0, top=1.0, bottom=0.0, left=0.0, right=1.0
    )
    tr = []
    for j in range(ny):
        tr.append([])
        for i in range(nx):
            left = i * wf * fudge_x
            bottom = fudge_y * (1.0 - (j + 1) * hf) + (1.0 - fudge_y)
            ax = fig.add_axes([left, bottom, wf * fudge_x, hf * fudge_y])
            tr[-1].append(ax)
    cbars = []
    if colorbar is None:
        pass
    elif colorbar.lower() == "horizontal":
        for i in range(nx):
            # left, bottom, width, height
            # Here we want 0.10 on each side of the colorbar
            # We want it to be 0.05 tall
            # And we want a buffer of 0.15
            ax = fig.add_axes(
                [
                    wf * (i + 0.10) * fudge_x,
                    hf * fudge_y * 0.20,
                    wf * (1 - 0.20) * fudge_x,
                    hf * fudge_y * 0.05,
                ]
            )
            cbars.append(ax)
    elif colorbar.lower() == "vertical":
        for j in range(ny):
            ax = fig.add_axes(
                [
                    wf * (nx + 0.05) * fudge_x,
                    hf * fudge_y * (ny - (j + 0.95)),
                    wf * fudge_x * 0.05,
                    hf * fudge_y * 0.90,
                ]
            )
            ax.clear()
            cbars.append(ax)
    return fig, tr, cbars



cdef struct IntegrationAccumulator:
    np.float64_t *t
    np.float64_t *dt
    np.uint8_t *child_mask
    int hits

cdef void dt_sampler(
             VolumeContainer *vc,
             np.float64_t v_pos[3],
             np.float64_t v_dir[3],
             np.float64_t enter_t,
             np.float64_t exit_t,
             int index[3],
             void *data) nogil:
    cdef IntegrationAccumulator *am = <IntegrationAccumulator *> data
    cdef int di = (index[0]*vc.dims[1]+index[1])*vc.dims[2]+index[2]
    if am.child_mask[di] == 0 or enter_t == exit_t:
        return
    am.hits += 1
    am.t[di] = enter_t
    am.dt[di] = (exit_t - enter_t)

cdef class RaySelector(SelectorObject):

    cdef public np.float64_t p1[3]
    cdef public np.float64_t p2[3]
    cdef public np.float64_t vec[3]

    def __init__(self, dobj):
        cdef int i
        _ensure_code(dobj.start_point)
        _ensure_code(dobj.end_point)
        for i in range(3):
            self.vec[i] = dobj.vec[i]
            self.p1[i] = dobj.start_point[i]
            self.p2[i] = dobj.end_point[i]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_mask(self, gobj):
        cdef np.ndarray[np.float64_t, ndim=3] t, dt
        cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
        cdef int i
        cdef int total = 0
        cdef IntegrationAccumulator *ia
        ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))
        cdef VolumeContainer vc
        mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
        t = np.zeros(gobj.ActiveDimensions, dtype="float64")
        dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1
        child_mask = gobj.child_mask
        ia.t = <np.float64_t *> t.data
        ia.dt = <np.float64_t *> dt.data
        ia.child_mask = <np.uint8_t *> child_mask.data
        ia.hits = 0
        _ensure_code(gobj.LeftEdge)
        _ensure_code(gobj.RightEdge)
        _ensure_code(gobj.dds)
        for i in range(3):
            vc.left_edge[i] = gobj.LeftEdge[i]
            vc.right_edge[i] = gobj.RightEdge[i]
            vc.dds[i] = gobj.dds[i]
            vc.idds[i] = 1.0/gobj.dds[i]
            vc.dims[i] = dt.shape[i]
        walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
        for i in range(dt.shape[0]):
            for j in range(dt.shape[1]):
                for k in range(dt.shape[2]):
                    if dt[i, j, k] >= 0:
                        mask[i, j, k] = 1
                        total += 1
        free(ia)
        if total == 0: return None
        return mask.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def get_dt(self, gobj):
        cdef np.ndarray[np.float64_t, ndim=3] t, dt
        cdef np.ndarray[np.float64_t, ndim=1] tr, dtr
        cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
        cdef int i, j, k, ni
        cdef IntegrationAccumulator *ia
        ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))
        cdef VolumeContainer vc
        t = np.zeros(gobj.ActiveDimensions, dtype="float64")
        dt = np.zeros(gobj.ActiveDimensions, dtype="float64") - 1
        child_mask = gobj.child_mask
        ia.t = <np.float64_t *> t.data
        ia.dt = <np.float64_t *> dt.data
        ia.child_mask = <np.uint8_t *> child_mask.data
        ia.hits = 0
        _ensure_code(gobj.LeftEdge)
        _ensure_code(gobj.RightEdge)
        _ensure_code(gobj.dds)
        for i in range(3):
            vc.left_edge[i] = gobj.LeftEdge[i]
            vc.right_edge[i] = gobj.RightEdge[i]
            vc.dds[i] = gobj.dds[i]
            vc.idds[i] = 1.0/gobj.dds[i]
            vc.dims[i] = dt.shape[i]
        walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
        tr = np.zeros(ia.hits, dtype="float64")
        dtr = np.zeros(ia.hits, dtype="float64")
        ni = 0
        for i in range(dt.shape[0]):
            for j in range(dt.shape[1]):
                for k in range(dt.shape[2]):
                    if dt[i, j, k] >= 0:
                        tr[ni] = t[i, j, k]
                        dtr[ni] = dt[i, j, k]
                        ni += 1
        if not (ni == ia.hits):
            print(ni, ia.hits)
        free(ia)
        return dtr, tr

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def get_dt_mesh(self, mesh, nz, int offset):
        cdef np.ndarray[np.float64_t, ndim=3] t, dt
        cdef np.ndarray[np.float64_t, ndim=1] tr, dtr
        cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
        cdef int i, j, k, ni
        cdef np.float64_t LE[3]
        cdef np.float64_t RE[3]
        cdef np.float64_t pos
        cdef IntegrationAccumulator *ia
        ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))
        cdef np.ndarray[np.float64_t, ndim=2] coords
        cdef np.ndarray[np.int64_t, ndim=2] indices
        indices = mesh.connectivity_indices
        coords = _ensure_code(mesh.connectivity_coords)
        cdef int nc = indices.shape[0]
        cdef int nv = indices.shape[1]
        if nv != 8:
            raise NotImplementedError
        cdef VolumeContainer vc
        child_mask = np.ones((1,1,1), dtype="uint8")
        t = np.zeros((1,1,1), dtype="float64")
        dt = np.zeros((1,1,1), dtype="float64") - 1
        tr = np.zeros(nz, dtype="float64")
        dtr = np.zeros(nz, dtype="float64")
        ia.t = <np.float64_t *> t.data
        ia.dt = <np.float64_t *> dt.data
        ia.child_mask = <np.uint8_t *> child_mask.data
        ia.hits = 0
        ni = 0
        for i in range(nc):
            for j in range(3):
                LE[j] = 1e60
                RE[j] = -1e60
            for j in range(nv):
                for k in range(3):
                    pos = coords[indices[i, j] - offset, k]
                    LE[k] = fmin(pos, LE[k])
                    RE[k] = fmax(pos, RE[k])
            for j in range(3):
                vc.left_edge[j] = LE[j]
                vc.right_edge[j] = RE[j]
                vc.dds[j] = RE[j] - LE[j]
                vc.idds[j] = 1.0/vc.dds[j]
                vc.dims[j] = 1
            t[0,0,0] = dt[0,0,0] = -1
            walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
            if dt[0,0,0] >= 0:
                tr[ni] = t[0,0,0]
                dtr[ni] = dt[0,0,0]
                ni += 1
        free(ia)
        return dtr, tr

    cdef int select_point(self, np.float64_t pos[3]) nogil:
        # two 0-volume constructs don't intersect
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:

        cdef int i
        cdef np.float64_t length = norm(self.vec)
        cdef np.float64_t r[3]
        for i in range(3):
            r[i] = pos[i] - self.p1[i]
        # the projected position of the sphere along the ray
        cdef np.float64_t l = dot(r, self.vec) / length
        # the square of the impact parameter
        cdef np.float64_t b_sqr = dot(r, r) - l*l

        # only accept spheres with radii larger than the impact parameter and
        # with a projected position along the ray no more than a radius away
        # from the ray
        if -radius < l and l < (length+radius) and b_sqr < radius*radius:
            return 1

        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        cdef int i, rv
        cdef VolumeContainer vc
        cdef IntegrationAccumulator *ia
        ia = <IntegrationAccumulator *> malloc(sizeof(IntegrationAccumulator))
        cdef np.float64_t dt[1]
        cdef np.float64_t t[1]
        cdef np.uint8_t cm[1]
        for i in range(3):
            vc.left_edge[i] = left_edge[i]
            vc.right_edge[i] = right_edge[i]
            vc.dds[i] = right_edge[i] - left_edge[i]
            vc.idds[i] = 1.0/vc.dds[i]
            vc.dims[i] = 1
        t[0] = dt[0] = 0.0
        cm[0] = 1
        ia.t = t
        ia.dt = dt
        ia.child_mask = cm
        ia.hits = 0
        walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> ia)
        rv = 0
        if ia.hits > 0:
            rv = 1
        free(ia)
        return rv

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        cdef int i
        cdef np.uint8_t cm = 1
        cdef VolumeContainer vc
        cdef IntegrationAccumulator ia
        cdef np.float64_t dt, t
        for i in range(3):
            vc.left_edge[i] = left_edge[i]
            vc.right_edge[i] = right_edge[i]
            vc.dds[i] = right_edge[i] - left_edge[i]
            vc.idds[i] = 1.0/vc.dds[i]
            vc.dims[i] = 1
        t = dt = 0.0
        ia.t = &t
        ia.dt = &dt
        ia.child_mask = &cm
        ia.hits = 0
        walk_volume(&vc, self.p1, self.vec, dt_sampler, <void*> &ia)
        if ia.hits > 0:
            return 2 # a box of non-zero volume cannot be inside a ray
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3],
                               np.float64_t dds[3]) nogil:
        # This is terribly inefficient for Octrees.  For grids, it will never
        # get called.
        cdef int i
        cdef np.float64_t left_edge[3]
        cdef np.float64_t right_edge[3]
        for i in range(3):
            left_edge[i] = pos[i] - dds[i]/2.0
            right_edge[i] = pos[i] + dds[i]/2.0
        return self.select_bbox(left_edge, right_edge)

    def _hash_vals(self):
        return (("p1[0]", self.p1[0]),
                ("p1[1]", self.p1[1]),
                ("p1[2]", self.p1[2]),
                ("p2[0]", self.p2[0]),
                ("p2[1]", self.p2[1]),
                ("p2[2]", self.p2[2]),
                ("vec[0]", self.vec[0]),
                ("vec[1]", self.vec[1]),
                ("vec[2]", self.vec[2]))

    def _get_state_attnames(self):
        return ("p1", "p2", "vec")

ray_selector = RaySelector



from yt.fields.field_info_container import FieldInfoContainer

# Common fields in FLASH: (Thanks to John ZuHone for this list)
#
# dens gas mass density (g/cc) --
# eint internal energy (ergs/g) --
# ener total energy (ergs/g), with 0.5*v^2 --
# gamc gamma defined as ratio of specific heats, no units
# game gamma defined as in , no units
# gpol gravitational potential from the last timestep (ergs/g)
# gpot gravitational potential from the current timestep (ergs/g)
# grac gravitational acceleration from the current timestep (cm s^-2)
# pden particle mass density (usually dark matter) (g/cc)
# pres pressure (erg/cc)
# temp temperature (K) --
# velx velocity x (cm/s) --
# vely velocity y (cm/s) --
# velz velocity z (cm/s) --

b_units = "code_magnetic"
pres_units = "code_mass/(code_length*code_time**2)"
en_units = "code_mass * (code_length/code_time)**2"
rho_units = "code_mass / code_length**3"


class FLASHFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("velx", ("code_length/code_time", ["velocity_x"], None)),
        ("vely", ("code_length/code_time", ["velocity_y"], None)),
        ("velz", ("code_length/code_time", ["velocity_z"], None)),
        ("dens", ("code_mass/code_length**3", ["density"], None)),
        ("temp", ("code_temperature", ["temperature"], None)),
        ("pres", (pres_units, ["pressure"], None)),
        ("gpot", ("code_length**2/code_time**2", ["gravitational_potential"], None)),
        ("gpol", ("code_length**2/code_time**2", [], None)),
        ("tion", ("code_temperature", [], None)),
        ("tele", ("code_temperature", [], None)),
        ("trad", ("code_temperature", [], None)),
        ("pion", (pres_units, [], None)),
        ("pele", (pres_units, [], "Electron Pressure, P_e")),
        ("prad", (pres_units, [], "Radiation Pressure")),
        ("eion", (en_units, [], "Ion Internal Specific Energy")),
        ("eele", (en_units, [], "Electron Internal Specific Energy")),
        ("erad", (en_units, [], "Radiation Internal Specific Energy")),
        ("pden", (rho_units, [], "Particle Mass Density")),
        ("depo", ("code_length**2/code_time**2", [], None)),
        ("ye", ("", [], "Y_e")),
        ("magp", (pres_units, [], None)),
        ("divb", ("code_magnetic/code_length", [], None)),
        ("game", ("", [], r"\gamma_e\ \rm{(ratio\ of\ specific\ heats)}")),
        ("gamc", ("", [], r"\gamma_c\ \rm{(ratio\ of\ specific\ heats)}")),
        ("flam", ("", [], None)),
        ("absr", ("", [], "Absorption Coefficient")),
        ("emis", ("", [], "Emissivity")),
        ("cond", ("", [], "Conductivity")),
        ("dfcf", ("", [], "Diffusion Equation Scalar")),
        ("fllm", ("", [], "Flux Limit")),
        ("pipe", ("", [], "P_i/P_e")),
        ("tite", ("", [], "T_i/T_e")),
        ("dbgs", ("", [], "Debug for Shocks")),
        ("cham", ("", [], "Chamber Material Fraction")),
        ("targ", ("", [], "Target Material Fraction")),
        ("sumy", ("", [], None)),
        ("mgdc", ("", [], "Emission Minus Absorption Diffusion Terms")),
        ("magx", (b_units, [], "B_x")),
        ("magy", (b_units, [], "B_y")),
        ("magz", (b_units, [], "B_z")),
    )

    known_particle_fields = (
        ("particle_posx", ("code_length", ["particle_position_x"], None)),
        ("particle_posy", ("code_length", ["particle_position_y"], None)),
        ("particle_posz", ("code_length", ["particle_position_z"], None)),
        ("particle_velx", ("code_length/code_time", ["particle_velocity_x"], None)),
        ("particle_vely", ("code_length/code_time", ["particle_velocity_y"], None)),
        ("particle_velz", ("code_length/code_time", ["particle_velocity_z"], None)),
        ("particle_tag", ("", ["particle_index"], None)),
        ("particle_mass", ("code_mass", ["particle_mass"], None)),
        (
            "particle_gpot",
            ("code_length**2/code_time**2", ["particle_gravitational_potential"], None),
        ),
    )

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        unit_system = self.ds.unit_system
        # Adopt FLASH 4.6 value for Na
        Na = self.ds.quan(6.022140857e23, "g**-1")
        for i in range(1, 1000):
            self.add_output_field(
                ("flash", f"r{i:03}"),
                sampling_type="cell",
                units="",
                display_name=f"Energy Group {i}",
            )

        # Add energy fields
        def ekin(data):
            ek = data["flash", "velx"] ** 2
            if data.ds.dimensionality >= 2:
                ek += data["flash", "vely"] ** 2
            if data.ds.dimensionality == 3:
                ek += data["flash", "velz"] ** 2
            return 0.5 * ek

        if ("flash", "ener") in self.field_list:
            self.add_output_field(
                ("flash", "ener"),
                sampling_type="cell",
                units="code_length**2/code_time**2",
            )
            self.alias(
                ("gas", "specific_total_energy"),
                ("flash", "ener"),
                units=unit_system["specific_energy"],
            )

        else:

            def _ener(field, data):
                ener = data["flash", "eint"] + ekin(data)
                try:
                    ener += data["flash", "magp"] / data["flash", "dens"]
                except Exception:
                    pass
                return ener

            self.add_field(
                ("gas", "specific_total_energy"),
                sampling_type="cell",
                function=_ener,
                units=unit_system["specific_energy"],
            )
        if ("flash", "eint") in self.field_list:
            self.add_output_field(
                ("flash", "eint"),
                sampling_type="cell",
                units="code_length**2/code_time**2",
            )
            self.alias(
                ("gas", "specific_thermal_energy"),
                ("flash", "eint"),
                units=unit_system["specific_energy"],
            )
        else:

            def _eint(field, data):
                eint = data["flash", "ener"] - ekin(data)
                try:
                    eint -= data["flash", "magp"] / data["flash", "dens"]
                except Exception:
                    pass
                return eint

            self.add_field(
                ("gas", "specific_thermal_energy"),
                sampling_type="cell",
                function=_eint,
                units=unit_system["specific_energy"],
            )

        ## Derived FLASH Fields

        if ("flash", "abar") in self.field_list:
            self.alias(("gas", "mean_molecular_weight"), ("flash", "abar"))
        elif ("flash", "sumy") in self.field_list:

            def _abar(field, data):
                return 1.0 / data["flash", "sumy"]

            self.add_field(
                ("gas", "mean_molecular_weight"),
                sampling_type="cell",
                function=_abar,
                units="",
            )
        elif "eos_singlespeciesa" in self.ds.parameters:

            def _abar(field, data):
                return data.ds.parameters["eos_singlespeciesa"] * data["index", "ones"]

            self.add_field(
                ("gas", "mean_molecular_weight"),
                sampling_type="cell",
                function=_abar,
                units="",
            )

        if ("flash", "sumy") in self.field_list:

            def _nele(field, data):
                return data["flash", "dens"] * data["flash", "ye"] * Na

            self.add_field(
                ("gas", "El_number_density"),
                sampling_type="cell",
                function=_nele,
                units=unit_system["number_density"],
            )

            def _nion(field, data):
                return data["flash", "dens"] * data["flash", "sumy"] * Na

            self.add_field(
                ("gas", "ion_number_density"),
                sampling_type="cell",
                function=_nion,
                units=unit_system["number_density"],
            )

            def _number_density(field, data):
                return (
                    data["gas", "El_number_density"] + data["gas", "ion_number_density"]
                )

        else:

            def _number_density(field, data):
                return data["flash", "dens"] * Na / data["gas", "mean_molecular_weight"]

        self.add_field(
            ("gas", "number_density"),
            sampling_type="cell",
            function=_number_density,
            units=unit_system["number_density"],
        )

        setup_magnetic_field_aliases(self, "flash", [f"mag{ax}" for ax in "xyz"])



import numpy as np

from yt.testing import assert_allclose_units, fake_random_ds, requires_file
from yt.units import cm, s
from yt.utilities.answer_testing.framework import data_dir_load
from yt.visualization.volume_rendering.off_axis_projection import off_axis_projection


def random_unit_vector(prng):
    v = prng.random_sample(3)
    while (v == 0).all():
        v = prng.random_sample(3)
    return v / np.sqrt((v ** 2).sum())


def random_velocity_vector(prng):
    return 2e5 * prng.random_sample(3) - 1e5


def compare_vector_conversions(data_source):
    prng = np.random.RandomState(8675309)
    normals = [[1, 0, 0], [0, 1, 0], [0, 0, 1]] + [
        random_unit_vector(prng) for i in range(2)
    ]
    bulk_velocities = [random_velocity_vector(prng) for i in range(2)]

    for bv in bulk_velocities:
        bulk_velocity = bv * cm / s
        data_source.set_field_parameter("bulk_velocity", bulk_velocity)
        data_source.clear_data()

        vmag = data_source[("gas", "velocity_magnitude")]
        vrad = data_source[("gas", "velocity_spherical_radius")]

        for normal in normals:
            data_source.set_field_parameter("normal", normal)
            data_source.clear_data()

            assert_allclose_units(
                vrad, data_source[("gas", "velocity_spherical_radius")]
            )

            vmag_new = data_source[("gas", "velocity_magnitude")]
            assert_allclose_units(vmag, vmag_new)

            vmag_cart = np.sqrt(
                (data_source[("gas", "velocity_x")] - bulk_velocity[0]) ** 2
                + (data_source[("gas", "velocity_y")] - bulk_velocity[1]) ** 2
                + (data_source[("gas", "velocity_z")] - bulk_velocity[2]) ** 2
            )
            assert_allclose_units(vmag, vmag_cart)

            vmag_cyl = np.sqrt(
                data_source[("gas", "velocity_cylindrical_radius")] ** 2
                + data_source[("gas", "velocity_cylindrical_theta")] ** 2
                + data_source[("gas", "velocity_cylindrical_z")] ** 2
            )
            assert_allclose_units(vmag, vmag_cyl)

            vmag_sph = np.sqrt(
                data_source[("gas", "velocity_spherical_radius")] ** 2
                + data_source[("gas", "velocity_spherical_theta")] ** 2
                + data_source[("gas", "velocity_spherical_phi")] ** 2
            )
            assert_allclose_units(vmag, vmag_sph)

            for i, d in enumerate("xyz"):
                assert_allclose_units(
                    data_source[("gas", f"velocity_{d}")] - bulk_velocity[i],
                    data_source[("gas", f"relative_velocity_{d}")],
                )

        for i, ax in enumerate("xyz"):
            data_source.set_field_parameter("axis", i)
            data_source.clear_data()
            assert_allclose_units(
                data_source[("gas", "velocity_los")],
                data_source[("gas", f"relative_velocity_{ax}")],
            )

        for i, ax in enumerate("xyz"):
            prj = data_source.ds.proj(
                ("gas", "velocity_los"), i, weight_field=("gas", "density")
            )
            assert_allclose_units(
                prj[("gas", "velocity_los")], prj[("gas", f"velocity_{ax}")]
            )

        data_source.clear_data()
        ax = [0.1, 0.2, -0.3]
        data_source.set_field_parameter("axis", ax)
        ax /= np.sqrt(np.dot(ax, ax))
        vlos = data_source[("gas", "relative_velocity_x")] * ax[0]
        vlos += data_source[("gas", "relative_velocity_y")] * ax[1]
        vlos += data_source[("gas", "relative_velocity_z")] * ax[2]
        assert_allclose_units(data_source[("gas", "velocity_los")], vlos)

        buf_los = off_axis_projection(
            data_source,
            data_source.ds.domain_center,
            ax,
            0.5,
            128,
            ("gas", "velocity_los"),
            weight=("gas", "density"),
        )

        buf_x = off_axis_projection(
            data_source,
            data_source.ds.domain_center,
            ax,
            0.5,
            128,
            ("gas", "relative_velocity_x"),
            weight=("gas", "density"),
        )

        buf_y = off_axis_projection(
            data_source,
            data_source.ds.domain_center,
            ax,
            0.5,
            128,
            ("gas", "relative_velocity_y"),
            weight=("gas", "density"),
        )

        buf_z = off_axis_projection(
            data_source,
            data_source.ds.domain_center,
            ax,
            0.5,
            128,
            ("gas", "relative_velocity_z"),
            weight=("gas", "density"),
        )

        vlos = buf_x * ax[0] + buf_y * ax[1] + buf_z * ax[2]

        assert_allclose_units(buf_los, vlos, rtol=1.0e-6)


def test_vector_component_conversions_fake():
    ds = fake_random_ds(16)
    ad = ds.all_data()
    compare_vector_conversions(ad)


g30 = "IsolatedGalaxy/galaxy0030/galaxy0030"


@requires_file(g30)
def test_vector_component_conversions_real():
    ds = data_dir_load(g30)
    sp = ds.sphere(ds.domain_center, (10, "kpc"))
    compare_vector_conversions(sp)



from .cartesian_coordinates import CartesianCoordinateHandler
from .coordinate_handler import _get_coord_fields


class SpectralCubeCoordinateHandler(CartesianCoordinateHandler):
    name = "spectral_cube"

    def __init__(self, ds, ordering=("x", "y", "z")):
        ordering = tuple(
            "xyz"[axis] for axis in (ds.lon_axis, ds.lat_axis, ds.spec_axis)
        )
        super().__init__(ds, ordering)

        self.default_unit_label = {}
        names = {}
        if ds.lon_name != "X" or ds.lat_name != "Y":
            names["x"] = r"Image\ x"
            names["y"] = r"Image\ y"
            # We can just use ds.lon_axis here
            self.default_unit_label[ds.lon_axis] = "pixel"
            self.default_unit_label[ds.lat_axis] = "pixel"
        names["z"] = ds.spec_name
        # Again, can use spec_axis here
        self.default_unit_label[ds.spec_axis] = ds.spec_unit

        self._image_axis_name = ian = {}
        for ax in "xyz":
            axi = self.axis_id[ax]
            xax = self.axis_name[self.x_axis[ax]]
            yax = self.axis_name[self.y_axis[ax]]
            ian[axi] = ian[ax] = ian[ax.upper()] = (
                names.get(xax, xax),
                names.get(yax, yax),
            )

        def _spec_axis(ax, x, y):
            p = (x, y)[ax]
            return [self.ds.pixel2spec(pp).v for pp in p]

        self.axis_field = {}
        self.axis_field[self.ds.spec_axis] = _spec_axis

    def setup_fields(self, registry):
        if not self.ds.no_cgs_equiv_length:
            return super().setup_fields(registry)
        for axi, ax in enumerate("xyz"):
            f1, f2 = _get_coord_fields(axi)

            def _get_length_func():
                def _length_func(field, data):
                    # Just use axis 0
                    rv = data.ds.arr(data.fcoords[..., 0].copy(), field.units)
                    rv[:] = 1.0
                    return rv

                return _length_func

            registry.add_field(
                ("index", f"d{ax}"),
                sampling_type="cell",
                function=f1,
                display_field=False,
                units="code_length",
            )

            registry.add_field(
                ("index", f"path_element_{ax}"),
                sampling_type="cell",
                function=_get_length_func(),
                display_field=False,
                units="",
            )

            registry.add_field(
                ("index", f"{ax}"),
                sampling_type="cell",
                function=f2,
                display_field=False,
                units="code_length",
            )

        def _cell_volume(field, data):
            rv = data["index", "dx"].copy(order="K")
            rv *= data["index", "dy"]
            rv *= data["index", "dz"]
            return rv

        registry.add_field(
            ("index", "cell_volume"),
            sampling_type="cell",
            function=_cell_volume,
            display_field=False,
            units="code_length**3",
        )
        registry.alias(("index", "volume"), ("index", "cell_volume"))

        registry.check_derived_fields(
            [
                ("index", "dx"),
                ("index", "dy"),
                ("index", "dz"),
                ("index", "x"),
                ("index", "y"),
                ("index", "z"),
                ("index", "cell_volume"),
            ]
        )

    _x_pairs = (("x", "y"), ("y", "x"), ("z", "x"))
    _y_pairs = (("x", "z"), ("y", "z"), ("z", "y"))

    def convert_to_cylindrical(self, coord):
        raise NotImplementedError

    def convert_from_cylindrical(self, coord):
        raise NotImplementedError

    @property
    def image_axis_name(self):
        return self._image_axis_name



from yt.fields.field_info_container import FieldInfoContainer
from yt.fields.magnetic_field import setup_magnetic_field_aliases
from yt.fields.species_fields import add_species_field_by_density, setup_species_fields
from yt.frontends.gadget.fields import GadgetFieldInfo
from yt.frontends.sph.fields import SPHFieldInfo

metal_elements = ["He", "C", "N", "O", "Ne", "Mg", "Si", "S", "Ca", "Fe"]


class GizmoFieldInfo(GadgetFieldInfo):
    # The known fields list is according to the GIZMO User Guide. See
    # http://www.tapir.caltech.edu/~phopkins/Site/GIZMO_files/gizmo_documentation.html#snaps-reading
    known_particle_fields = (
        ("Coordinates", ("code_length", ["particle_position"], None)),
        ("Velocities", ("code_velocity", ["particle_velocity"], None)),
        ("ParticleIDs", ("", ["particle_index"], None)),
        ("Masses", ("code_mass", ["particle_mass"], None)),
        ("InternalEnergy", ("code_specific_energy", ["specific_thermal_energy"], None)),
        ("Density", ("code_mass / code_length**3", ["density"], None)),
        ("SmoothingLength", ("code_length", ["smoothing_length"], None)),
        ("ElectronAbundance", ("", [], None)),
        ("NeutralHydrogenAbundance", ("", [], None)),
        ("StarFormationRate", ("Msun / yr", [], None)),
        ("Metallicity", ("code_metallicity", ["metallicity"], None)),
        ("Metallicity_00", ("", ["metallicity"], None)),
        ("Metallicity_01", ("", ["He_metallicity"], None)),
        ("Metallicity_02", ("", ["C_metallicity"], None)),
        ("Metallicity_03", ("", ["N_metallicity"], None)),
        ("Metallicity_04", ("", ["O_metallicity"], None)),
        ("Metallicity_05", ("", ["Ne_metallicity"], None)),
        ("Metallicity_06", ("", ["Mg_metallicity"], None)),
        ("Metallicity_07", ("", ["Si_metallicity"], None)),
        ("Metallicity_08", ("", ["S_metallicity"], None)),
        ("Metallicity_09", ("", ["Ca_metallicity"], None)),
        ("Metallicity_10", ("", ["Fe_metallicity"], None)),
        ("ArtificialViscosity", ("", [], None)),
        ("MagneticField", ("code_magnetic", ["particle_magnetic_field"], None)),
        ("DivergenceOfMagneticField", ("code_magnetic / code_length", [], None)),
        ("StellarFormationTime", ("", [], None)),
        # "StellarFormationTime" has different meanings in (non-)cosmological
        # runs, so units are left blank here.
        ("BH_Mass", ("code_mass", [], None)),
        ("BH_Mdot", ("code_mass / code_time", [], None)),
        ("BH_Mass_AlphaDisk", ("code_mass", [], None)),
    )

    def __init__(self, *args, **kwargs):
        super(SPHFieldInfo, self).__init__(*args, **kwargs)
        if ("PartType0", "Metallicity_00") in self.field_list:
            self.nuclei_names = metal_elements
            self.species_names = ["H_p0", "H_p1"] + metal_elements

    def setup_particle_fields(self, ptype):
        FieldInfoContainer.setup_particle_fields(self, ptype)
        if ptype in ("PartType0",):
            self.setup_gas_particle_fields(ptype)
            setup_species_fields(self, ptype)
        if ptype in ("PartType4",):
            self.setup_star_particle_fields(ptype)

    def setup_gas_particle_fields(self, ptype):
        super().setup_gas_particle_fields(ptype)

        def _h_p0_density(field, data):
            x_H = 1.0 - data[(ptype, "He_metallicity")] - data[(ptype, "metallicity")]
            return (
                x_H
                * data[(ptype, "density")]
                * data[(ptype, "NeutralHydrogenAbundance")]
            )

        self.add_field(
            (ptype, "H_p0_density"),
            sampling_type="particle",
            function=_h_p0_density,
            units=self.ds.unit_system["density"],
        )
        add_species_field_by_density(self, ptype, "H")

        def _h_p1_density(field, data):
            x_H = 1.0 - data[(ptype, "He_metallicity")] - data[(ptype, "metallicity")]
            return (
                x_H
                * data[(ptype, "density")]
                * (1.0 - data[(ptype, "NeutralHydrogenAbundance")])
            )

        self.add_field(
            (ptype, "H_p1_density"),
            sampling_type="particle",
            function=_h_p1_density,
            units=self.ds.unit_system["density"],
        )
        add_species_field_by_density(self, ptype, "H_p1")

        def _nuclei_mass_density_field(field, data):
            species = field.name[1][: field.name[1].find("_")]
            return data[ptype, "density"] * data[ptype, f"{species}_metallicity"]

        for species in ["H", "H_p0", "H_p1"]:
            for suf in ["_density", "_number_density"]:
                field = f"{species}{suf}"
                self.alias(("gas", field), (ptype, field))

        if (ptype, "ElectronAbundance") in self.field_list:

            def _el_number_density(field, data):
                return (
                    data[ptype, "ElectronAbundance"] * data[ptype, "H_nuclei_density"]
                )

            self.add_field(
                (ptype, "El_number_density"),
                sampling_type="particle",
                function=_el_number_density,
                units=self.ds.unit_system["number_density"],
            )
            self.alias(("gas", "El_number_density"), (ptype, "El_number_density"))

        for species in self.nuclei_names:
            self.add_field(
                (ptype, f"{species}_nuclei_mass_density"),
                sampling_type="particle",
                function=_nuclei_mass_density_field,
                units=self.ds.unit_system["density"],
            )

            for suf in ["_nuclei_mass_density", "_metallicity"]:
                field = f"{species}{suf}"
                self.alias(("gas", field), (ptype, field))

        def _metal_density_field(field, data):
            return data[ptype, "metallicity"] * data[ptype, "density"]

        self.add_field(
            (ptype, "metal_density"),
            sampling_type="local",
            function=_metal_density_field,
            units=self.ds.unit_system["density"],
        )
        self.alias(("gas", "metal_density"), (ptype, "metal_density"))

        magnetic_field = "MagneticField"
        if (ptype, magnetic_field) in self.field_list:
            setup_magnetic_field_aliases(self, ptype, magnetic_field)

    def setup_star_particle_fields(self, ptype):
        def _creation_time(field, data):
            if data.ds.cosmological_simulation:
                a_form = data[(ptype, "StellarFormationTime")]
                z_form = 1 / a_form - 1
                creation_time = data.ds.cosmology.t_from_z(z_form)
            else:
                t_form = data[(ptype, "StellarFormationTime")]
                creation_time = data.ds.arr(t_form, "code_time")
            return creation_time

        self.add_field(
            (ptype, "creation_time"),
            sampling_type="particle",
            function=_creation_time,
            units=self.ds.unit_system["time"],
        )

        def _age(field, data):
            return data.ds.current_time - data[(ptype, "creation_time")]

        self.add_field(
            (ptype, "age"),
            sampling_type="particle",
            function=_age,
            units=self.ds.unit_system["time"],
        )
